/*
 * [CSVCondense.java]
 *
 * Summary: Remove blank lines and condense multiple spaces to one within fields. Trim lead and trail blanks on fields.
 *
 * Copyright: (c) 2011 Roedy Green, Canadian Mind Products, http://mindprod.com
 *
 * Licence: This software may be copied and used freely for any purpose but military.
 *          http://mindprod.com/contact/nonmil.html
 *
 * Requires: JDK 1.5+
 *
 * Created with: JetBrains IntelliJ IDEA IDE http://www.jetbrains.com/idea/
 *
 * Version History:
 *  1.0 2011-02-08 initial version
 *  1.1 2011-02-17 tidy up the label comment
 */
package com.mindprod.csv;

import com.mindprod.common15.STA;
import com.mindprod.fastcat.FastCat;
import com.mindprod.hunkio.HunkIO;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.EOFException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.util.regex.Pattern;

import static java.lang.System.err;
import static java.lang.System.out;

/**
 * Remove blank lines and condense multiple spaces to one within fields. Trim lead and trail blanks on fields.
 * <p/>
 * Use: java.exe com.mindprod.CSVCondense somefile.csv
 *
 * @author Roedy Green, Canadian Mind Products
 * @version 1.1 2011-02-17 tidy up the label comment
 * @since 2011-02-08
 */
public final class CSVCondense
    {
    // ------------------------------ CONSTANTS ------------------------------

    /**
     * how to use the command line
     */
    private static final String USAGE = "CSVCondense needs a single filename.csv on the command line.";

    /**
     * used to split Label comment fields apart
     */
    static final Pattern SPLITTER = Pattern.compile( "\\s*,\\s*" );

    // -------------------------- PUBLIC INSTANCE  METHODS --------------------------

    /**
     * Condense a CSV file, constructor. Just create. There are no methods to call.
     *
     * @param file          CSV file to be packed to remove excess space and quotes.
     * @param separatorChar field separator character, usually ',' in North America, ';' in Europe and sometimes '\t' for
     *                      tab.
     * @param quoteChar     char to use to enclose fields containing a separator, usually '\"'. Use (char)0 if
     *                      you don't want a quote character.
     * @param commentChar   char to use to introduce comments.  Use (char) 0 if none.  Only one character allowed.
     * @param encoding      encoding of input and output file.
     *
     * @throws java.io.IOException if problems reading/writing file
     */
    @SuppressWarnings( { "WeakerAccess" } )
    public CSVCondense( final File file, final char separatorChar, final char quoteChar, final char commentChar, final String encoding ) throws IOException
        {
        final String commentChars = ( commentChar == 0 ) ? "" : String.valueOf( commentChar );

        // reader, separatorChar, quoteChar, commentChars, hideComments, trimQuoted, trimpUnquoted, allowMultipleLineFields
        final CSVReader r = new CSVReader( new InputStreamReader( new BufferedInputStream( new FileInputStream( file ), 32768 ), encoding ),
                separatorChar, quoteChar, commentChars, false /* hideComments */, true /* trimQuoted */, true /* trimUnquoted */, true /* multiline */ );

        final File tempFile = HunkIO.createTempFile( "temp", ".tmp", file );
        // writer, quoteLevel, separatorChar, quoteChar, commentChar, trim
        final PrintWriter pw = new PrintWriter( new OutputStreamWriter( new BufferedOutputStream( new FileOutputStream( tempFile ), 32768 ), encoding ) );
        final CSVWriter w = new CSVWriter( pw, 0 /* minimal  */, separatorChar, quoteChar, commentChar, true );

        try
            {
            while ( true )
                {
                final String[] next = r.getAllFieldsInLine();
                if ( next.length != 0 )
                    {
                    final boolean wasLastFieldAComment = r.wasComment();
                    for ( int i = 0; i < next.length; i++ )
                        {
                        next[ i ] = STA.condense( next[ i ] ); // condense comment too.
                        }
                    if ( r.wasLabelComment() )
                        {
                        next[ next.length - 1 ] = rebuildLabelComment( next[ next.length - 1 ] );
                        }

                    w.nl( next, wasLastFieldAComment );
                    }
                }
            }

        catch ( EOFException e )
            {
            out.println( r.getLineCount() + " lines condensed." );
            r.close();
            w.close();
            file.delete();
            tempFile.renameTo( file );
            }
        }

    // -------------------------- STATIC METHODS --------------------------

    /**
     * comment has two lead comment chars, one stripped off, and are used to label the fields.
     * Reshape it so it labels the new fields.
     *
     * @param comment field names separated by commas, with lead comment char.
     *
     * @return reordered comment
     */
    static String rebuildLabelComment( String comment )
        {
        final String[] fieldnames = SPLITTER.split( comment.substring( 1 ) );  // strip second #
        final FastCat sb = new FastCat( fieldnames.length * 2 + 2 );
        sb.append( comment.charAt( 0 ) ); // rebuild just second #
        sb.append( ' ' );
        for ( String fieldname : fieldnames )
            {
            sb.append( STA.condense( fieldname ) );
            sb.append( "," );
            }

        return sb.toString().substring( 0, sb.length() - 1 ); // chop off final ,
        }

    // --------------------------- main() method ---------------------------

    /**
     * Simple command line interface to CSVCondense  Condenses csv file whose name is on the command line. Must have
     * extension .csv <br /> Use java com.mindprod.CSVCondense  somefile.csv
     *
     * @param args name of csv file to remove excess space and blank lines
     */
    public static void main( final String[] args )
        {
        if ( args.length < 1 )
            {
            throw new IllegalArgumentException( USAGE );
            }
        final String filename = args[ 0 ];
        if ( !filename.endsWith( ".csv" ) )
            {
            throw new IllegalArgumentException( "Bad Extension\n" + USAGE );
            }

        try
            {
            // file,  separatorChar,  quoteChar,  commentChar, cols that can be different
            new CSVCondense( new File( filename ), ',', '\"', '#', "UTF-8" );
            }
        catch ( IOException e )
            {
            err.println();
            e.printStackTrace( err );
            err.println( "CSVCondense failed to condense " + filename );
            err.println();
            }
        }// end main
    }
