/*
 * [CSVDeDup.java]
 *
 * Summary: DeDup CSV File: remove duplicate records.
 *
 * Copyright: (c) 2010-2011 Roedy Green, Canadian Mind Products, http://mindprod.com
 *
 * Licence: This software may be copied and used freely for any purpose but military.
 *          http://mindprod.com/contact/nonmil.html
 *
 * Requires: JDK 1.5+
 *
 * Created with: JetBrains IntelliJ IDEA IDE http://www.jetbrains.com/idea/
 *
 * Version History:
 *  1.0 2010-12-17 initial version
 *  1.1 2011-01-25 allow you to specify encoding
 */
package com.mindprod.csv;

import com.mindprod.hunkio.HunkIO;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.EOFException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;

import static java.lang.System.err;
import static java.lang.System.out;

/**
 * DeDup CSV File: remove duplicate records.
 * <p/>
 * Use: java.exe com.mindprod.CSVDeDup somefile.csv 2 3
 * may specify 0-based columns of column to exclude when determining dups. These columns are allowed to be
 * different, and if the other columns match, subsequent matching records are discarded.
 * Comments are ignored when determining duplicates.
 *
 * @author Roedy Green, Canadian Mind Products
 * @version 1.1 2011-01-25 allow you to specify encoding
 * @since 2010
 */
public final class CSVDeDup
    {
    // ------------------------------ CONSTANTS ------------------------------

    /**
     * how to use the command line
     */
    private static final String USAGE = "CSVDeDup needs a single filename.csv and exception columns on the command line.";

    // -------------------------- PUBLIC INSTANCE  METHODS --------------------------

    /**
     * DeDup a CSV file, constructor. Just create. There are no methods to call.
     *
     * @param file          CSV file to be packed to remove excess space and quotes.
     * @param separatorChar field separator character, usually ',' in North America, ';' in Europe and sometimes '\t' for
     *                      tab.
     * @param quoteChar     char to use to enclose fields containing a separator, usually '\"'. Use (char)0 if
     *                      you don't want a quote character.
     * @param commentChar   char to use to introduce comments.  Use (char) 0 if none.  Only one character allowed.
     * @param encoding      encoding of input and output file.
     * @param exceptionCols columns that are allowed to be different, and still count as a duplicate.
     *
     * @throws java.io.IOException if problems reading/writing file
     */
    @SuppressWarnings( { "WeakerAccess" } )
    public CSVDeDup( final File file, final char separatorChar, final char quoteChar, final char commentChar, final String encoding, final int... exceptionCols ) throws IOException
        {
        final String commentChars = ( commentChar == 0 ) ? "" : String.valueOf( commentChar );
        final boolean[] exclude = new boolean[ 100 ];
        for ( int exceptionCol : exceptionCols )
            {
            exclude[ exceptionCol ] = true;
            }

        // reader, separatorChar, quoteChar, commentChars, hideComments, trimQuoted, trimUnquoted, allowMultipleLineFields
        final CSVReader r = new CSVReader( new InputStreamReader( new BufferedInputStream( new FileInputStream( file ), 32768 ), encoding ),
                separatorChar, quoteChar, commentChars, false, true /* trimQuoted */, true /* trimUnquoted */, true );

        File tempFile = HunkIO.createTempFile( "temp", ".tmp", file );
        // writer, quoteLevel, separatorChar, quoteChar, commentChar, trim
        final PrintWriter pw = new PrintWriter( new OutputStreamWriter( new BufferedOutputStream( new FileOutputStream( tempFile ), 32768 ), encoding ) );
        final CSVWriter w = new CSVWriter( pw, 0 /* minimal  */, separatorChar, quoteChar, commentChar, true );

        String[] prev = new String[ 0 ];
        int prevLength = 0;
        try
            {
            while ( true )
                {
                final String[] next = r.getAllFieldsInLine();
                final boolean wasLastFieldAComment = r.wasComment();
                final int nextLength = next.length - ( wasLastFieldAComment ? 1 : 0 );
                final int fieldCount = Math.min( nextLength, prevLength );
                boolean different = false;

                if ( wasLastFieldAComment && nextLength == 0 )
                    {
                    // always keep col 0 comments
                    different = true;
                    }
                else
                    {
                    triplefor:
                    {
                    for ( int i = 0; i < fieldCount; i++ )
                        {
                        if ( !exclude[ i ] && !next[ i ].equals( prev[ i ] ) )
                            {
                            different = true;
                            break triplefor;
                            }
                        }
                    // handle part of next that is longer than prev, if any
                    for ( int i = fieldCount; i < nextLength; i++ )
                        {
                        if ( !exclude[ i ] && next[ i ].length() != 0 )
                            {
                            different = true;
                            break triplefor;
                            }
                        }
                    // handle part of prev that is longer than next, if any
                    for ( int i = fieldCount; i < prevLength; i++ )
                        {
                        if ( !exclude[ i ] && prev[ i ].length() != 0 )
                            {
                            different = true;
                            break triplefor;
                            }
                        }
                    }
                    }
                if ( different )
                    {
                    w.nl( next, wasLastFieldAComment );
                    prev = next;
                    prevLength = nextLength;
                    }
                }
            }
        catch ( EOFException e )
            {
            out.println( r.getLineCount() + " lines read, "
                         + w.getLineCount() + " lines written." );
            r.close();
            w.close();
            file.delete();
            tempFile.renameTo( file );
            }
        }

    // --------------------------- main() method ---------------------------

    /**
     * Simple command line interface to CSVDeDup. Dedups one csv file whose name is on the command line. Must have
     * extension .csv <br /> Use java com.mindprod.CSVDeDup somefile.csv 2  3
     *
     * @param args name of csv file to remove excess quotes and space
     */
    public static void main( final String[] args )
        {
        if ( args.length < 1 )
            {
            throw new IllegalArgumentException( USAGE );
            }
        final String filename = args[ 0 ];
        if ( !filename.endsWith( ".csv" ) )
            {
            throw new IllegalArgumentException( "Bad Extension\n" + USAGE );
            }
        final int[] exceptionCols = new int[ args.length - 1 ];
        for ( int i = 1; i < args.length; i++ )
            {
            exceptionCols[ i - 1 ] = Integer.parseInt( args[ i ] );
            }
        try
            {
            // file,  separatorChar,  quoteChar,  commentChar, cols that can be different
            new CSVDeDup( new File( filename ), ',', '\"', '#', "UTF-8", exceptionCols );
            }
        catch ( IOException e )
            {
            err.println();
            e.printStackTrace( err );
            err.println( "CSVDeDup failed to dedup " + filename );
            err.println();
            }
        }// end main
    }
