/*
 * [CSVPatch.java]
 *
 * Summary: patch a CSV file with a list of from:to pairs.
 *
 * Copyright: (c) 2010-2011 Roedy Green, Canadian Mind Products, http://mindprod.com
 *
 * Licence: This software may be copied and used freely for any purpose but military.
 *          http://mindprod.com/contact/nonmil.html
 *
 * Requires: JDK 1.5+
 *
 * Created with: JetBrains IntelliJ IDEA IDE http://www.jetbrains.com/idea/
 *
 * Version History:
 *  1.0 2010-12-30 initial version
 *  1.1 2011-01-25 allow you to specify encoding
 *  1.2 2011-03-08 eliminate all but first duplicate in patch file. allow comments in patch file, allow 2+ cols in patch file.
 */
package com.mindprod.csv;

import com.mindprod.hunkio.HunkIO;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.EOFException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.util.HashMap;

import static java.lang.System.err;
import static java.lang.System.out;

/**
 * patch a CSV file with a list of from:to pairs.
 * <p/>
 * Use: java.exe com.mindprod.CSVPatch somefile.csv  fromto.csv  0 3  (columns to patch)
 * Typically used to update URLs in a column with recently discovered replacements.
 *
 * @author Roedy Green, Canadian Mind Products
 * @version 1.2 2011-03-08 eliminate all but first duplicate in patch file. allow comments in patch file, allow 2+ cols in patch file.
 * @since 2010-12-30
 */
public final class CSVPatch
    {
    // ------------------------------ CONSTANTS ------------------------------

    /**
     * how to use the command line
     */
    private static final String USAGE = "CSVPatch needs filenametopatch.csv fromtopairsfilename.csv 0 3 ... (0-based cols to patch)";

    // -------------------------- PUBLIC INSTANCE  METHODS --------------------------

    /**
     * patch a CSV fileToPatch, constructor. Just create. There are no methods to call.
     *
     * @param fileToPatch   CSV fileToPatch to be packed to remove excess space and quotes.
     * @param fileOfPairs   CSV file of pairs from,to (extra cols ignored)
     * @param separatorChar field separator character, usually ',' in North America, ';' in Europe and sometimes '\t' for
     *                      tab.
     * @param quoteChar     char to use to enclose fields containing a separator, usually '\"'. Use (char)0 if
     *                      you don't want a quote character.
     * @param commentChar   char to use to introduce comments.  Use (char) 0 if none.  Only one character allowed.
     * @param encoding      encoding of input and output.
     * @param colsToPatch   list of columns that should be patched, 0-based.
     *
     * @throws java.io.IOException if problems reading/writing fileToPatch
     */
    @SuppressWarnings( { "WeakerAccess" } )
    public CSVPatch( final File fileToPatch,
                     final File fileOfPairs,
                     final char separatorChar,
                     final char quoteChar,
                     final char commentChar,
                     final String encoding,
                     final int... colsToPatch ) throws IOException
        {
        // hold translate from:to pairs in RAM
        final HashMap<String, String> translator = new HashMap<String, String>( 2048 );

        final String commentChars = ( commentChar == 0 ) ? "" : String.valueOf( commentChar );

        // reader, separatorChar, quoteChar, commentChars, hideComments, trimQuoted, trimUnquoted, allowMultipleLineFields
        final CSVReader inPairs = new CSVReader( new InputStreamReader( new BufferedInputStream( new FileInputStream( fileOfPairs ), 32768 ), encoding ), separatorChar, quoteChar, commentChars,
                true /* hide comments */, true /* trimQuoted */, true /* trimUnquoted */, true /* allow multi-line */ );
        int pairCount = 0;
        try
            {
            while ( true )
                {
                final String from = inPairs.get();
                // ignore blank line
                if ( from == null )
                    {
                    continue;
                    }
                final String to = inPairs.get();
                if ( to == null || from.length() == 0 )
                    {
                    throw new IllegalArgumentException( "pairs file " + fileOfPairs.toString() + " must contain at least two fields (from,to) on each line" );
                    }
                inPairs.skipToNextLine();
                if ( !translator.containsKey( from ) )
                    {
                    translator.put( from, to );
                    }
                }
            }
        catch ( EOFException e )
            {
            pairCount = inPairs.getLineCount();
            inPairs.close();
            }

        // reader, separatorChar, quoteChar, commentChars, hideComments, trimQuoted, allowMultipleLineFields
        final CSVReader r = new CSVReader( new InputStreamReader( new BufferedInputStream( new FileInputStream( fileToPatch ), 32768 ), encoding ),
                separatorChar, quoteChar, commentChars, false /* process comments */, true /* trimQuoted */, true /* trimUnquoted */, true /* allow multi-line */ );

        File tempFile = HunkIO.createTempFile( "temp", ".tmp", fileToPatch );
        // writer, quoteLevel, separatorChar, quoteChar, commentChar, trim
        final PrintWriter pw = new PrintWriter( new OutputStreamWriter( new BufferedOutputStream( new FileOutputStream( tempFile ), 32768 ), encoding ) );
        final CSVWriter w = new CSVWriter( pw, 0 /* minimal  */, separatorChar, quoteChar, commentChar, true /* trim */ );
        int fieldsPatched = 0;
        try
            {
            while ( true )
                {
                final String[] fields = r.getAllFieldsInLine();
                final boolean lastFieldWasComment = r.wasComment();
                // patch each of the columns requested.
                for ( int col : colsToPatch )
                    {
                    if ( col < fields.length )
                        {
                        final String newValue = translator.get( fields[ col ] );
                        if ( newValue != null )
                            {
                            fields[ col ] = newValue;
                            fieldsPatched++;
                            }
                        // most of the time we make no change. The value won't be in the HashMap.
                        }
                    w.nl( fields, lastFieldWasComment );
                    }
                }
            }
        catch ( EOFException e )
            {
            out.println( r.getLineCount() + " lines read, "
                         + pairCount + " pairs read, "
                         + fieldsPatched + " fields patched, "
                         + w.getLineCount() + " lines written." );

            r.close();
            w.close();
            fileToPatch.delete();
            tempFile.renameTo( fileToPatch );
            }
        }

    // --------------------------- main() method ---------------------------

    /**
     * Simple command line interface to CSVPatch. Patches one csv file whose name is on the command line with a list
     * of from to pairs in another csv file.  You specify which columns to translate from to to.
     * Both files must have the extension .csv
     *
     * @param args name of csv file to patch, from-to pairs file, columns to patch.
     */
    public static void main( String[] args )
        {
        if ( args.length < 3 )
            {
            throw new IllegalArgumentException(
                    USAGE );
            }
        final String fileToPatch = args[ 0 ];
        if ( !fileToPatch.endsWith( ".csv" ) )
            {
            throw new IllegalArgumentException(
                    "Bad filenametopatch Extension\n" + USAGE );
            }
        final String fileOfPairs = args[ 1 ];
        if ( !fileOfPairs.endsWith( ".csv" ) )
            {
            throw new IllegalArgumentException(
                    "Bad fromtopairs Extension\n" + USAGE );
            }
        final int[] colsToPatch = new int[ args.length - 2 ];
        try
            {
            for ( int i = 2; i < args.length; i++ )
                {
                colsToPatch[ i - 2 ] = Integer.parseInt( args[ i ] );
                }
            }
        catch ( NumberFormatException e )
            {
            throw new IllegalArgumentException( USAGE );
            }
        try
            {
            // file,  separatorChar,  quoteChar,  commentChar
            new CSVPatch( new File( fileToPatch ), new File( fileOfPairs ), ',', '\"', '#', "UTF-8", colsToPatch );
            }
        catch ( IOException e )
            {
            err.println();
            e.printStackTrace( err );
            err.println( "CSVPatch failed to patch " + fileToPatch );
            err.println();
            }
        }// end main
    }
