/*
 * [CSVToTable.java]
 *
 * Summary: Converts a CSV file to the guts of an HTML table. Output appears in xxx.html.
 *
 * Copyright: (c) 2011 Roedy Green, Canadian Mind Products, http://mindprod.com
 *
 * Licence: This software may be copied and used freely for any purpose but military.
 *          http://mindprod.com/contact/nonmil.html
 *
 * Requires: JDK 1.5+
 *
 * Created with: JetBrains IntelliJ IDEA IDE http://www.jetbrains.com/idea/
 *
 * Version History:
 *  1.0 2011-01-23 initial version.
 *  1.1 2011-01-25 allow optional css classes on command line, encoding.
 *  1.2 2011-02-14 no longer entify. Do separately with Entify.
 */
package com.mindprod.csv;

import com.mindprod.fastcat.FastCat;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.EOFException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;

import static java.lang.System.err;
import static java.lang.System.out;

/**
 * Converts a CSV file to the guts of an HTML table. Output appears in xxx.html.
 * <p/>
 * Use: java.exe com.mindprod.CSVToTable xxx.csv
 * Awkward characters will appear as Entities.
 *
 * @author Roedy Green, Canadian Mind Products
 * @version 1.2 2011-02-14 no longer entify. Do separately with Entify.
 * @since 2011-01-23
 */
public final class CSVToTable
    {
    // ------------------------------ CONSTANTS ------------------------------

    /**
     * how to use the command line
     */
    private static final String USAGE = "CSVToTable needs the name of a CSV file on the commandline,\nfollowed optionally by css classes for each column.\nOutput will be in xxx.html.";

    // -------------------------- PUBLIC INSTANCE  METHODS --------------------------

    /**
     * Constructor to convert a CSV file to an HTML table.
     *
     * @param file          CSV file to be packed to remove excess space and quotes.
     * @param separatorChar field separator character, usually ',' in North America, ';' in Europe and sometimes '\t' for
     *                      tab for the output file.  It is tab for the input file.
     *                      Note this is a 'char' not a "string".
     * @param quoteChar     character used to quote fields containing awkward chars.
     * @param commentChars  characters to treat as comments.
     * @param encoding      encoding of the input and output file.
     * @param cssClasses    css Classes for columns of HTML table (optional)
     *
     * @throws java.io.IOException if problems reading/writing file
     */
    @SuppressWarnings( { "WeakerAccess" } )
    public CSVToTable( final File file, final char separatorChar, final char quoteChar, final String commentChars, final String encoding, final String... cssClasses ) throws IOException
        {
        // reader, separatorChar, quoteChar, commentChars, hideComments, trimQuoted, trimUnquoted, allowMultipleLineFields
        final CSVReader r = new CSVReader( new InputStreamReader( new BufferedInputStream( new FileInputStream( file ), 32768 ), encoding ), separatorChar, quoteChar, commentChars,
                true, true /* trimQuoted */, true /* trimUnquoted */, true );

        String tableFilename = file.getAbsolutePath();
        // replace .csv with .srs
        tableFilename = tableFilename.substring( 0, tableFilename.length() - 4 ) + ".html";
        final File tableFile = new File( tableFilename );
        final PrintWriter w = new PrintWriter( new OutputStreamWriter( new BufferedOutputStream( new FileOutputStream( tableFile ), 16384 ), encoding ) );
        try
            {
            while ( true )
                {
                final String[] fields = r.getAllFieldsInLine();

                // we don't apple the <table or </table>
                final FastCat sb = new FastCat( fields.length * 5 + 2 );
                sb.append( "<tr>" );
                for ( int i = 0; i < fields.length; i++ )
                    {
                    final String field = fields[ i ];
                    if ( i < cssClasses.length && cssClasses[ i ].length() > 0 )
                        {
                        sb.append( "<td class=\"" );
                        sb.append( cssClasses[ i ] );
                        sb.append( "\">" );
                        }
                    else
                        {
                        sb.append( "<td>" );
                        }
                    sb.append( field );
                    sb.append( "</td>" );
                    }
                sb.append( "</tr>" );
                w.println( sb.toString() );
                }
            }
        catch ( EOFException e )
            {
            out.println( r.getLineCount() + " csv lines converted to table rows." );
            r.close();
            w.close();
            }
        }

    // --------------------------- main() method ---------------------------

    /**
     * Simple command line interface to CSVToTable. Converts one CSV file to an HTML table. Must have
     * extension .csv <br /> Use java com.mindprod.CSVToTable somefile.csv.
     * You may optionally provide CSS classes for each column of the table. "" means no css class for that column.
     * You can use CSVToTable constructor in your own programs.
     *
     * @param args name of csv file to remove excess quotes and space
     */
    public static void main( String[] args )
        {
        if ( args.length < 1 )
            {
            throw new IllegalArgumentException( USAGE );
            }
        String filename = args[ 0 ];
        if ( !filename.endsWith( ".csv" ) )
            {
            throw new IllegalArgumentException( "Bad Extension\n" + USAGE );
            }

        final String[] cssClasses = new String[ args.length - 1 ];
        System.arraycopy( args, 1, cssClasses, 0, args.length - 1 );
        try
            {
            // file,  separatorChar,  quoteChar,  commentChar
            new CSVToTable( new File( filename ), ',', '\"', "#", "UTF-8", cssClasses );
            }
        catch ( IOException e )
            {
            err.println();
            e.printStackTrace( err );
            err.println( "CSVToTable failed to export" + filename );
            err.println();
            }
        }// end main
    }
