/*
 * [LinesToCSV.java]
 *
 * Summary: Converts text file of one field per line to a comma-separated CSV file.
 *
 * Copyright: (c) 2008-2011 Roedy Green, Canadian Mind Products, http://mindprod.com
 *
 * Licence: This software may be copied and used freely for any purpose but military.
 *          http://mindprod.com/contact/nonmil.html
 *
 * Requires: JDK 1.5+
 *
 * Created with: JetBrains IntelliJ IDEA IDE http://www.jetbrains.com/idea/
 *
 * Version History:
 *  1.0 2010-12-14 initial version
 *  1.1 2011-01-25 allow you to specify encoding
 */
package com.mindprod.csv;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.BufferedReader;
import java.io.EOFException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;

import static java.lang.System.err;
import static java.lang.System.out;

/**
 * Converts text file of one field per line to a comma-separated CSV file.
 * <p/>
 * Use: java.exe com.mindprod.LinesToCSV somefile.txt  3
 * How many lines should be grouped together on one line of the CSV file.
 * Output is somefile.csv.  somefile.txt is left unchanged.
 *
 * @author Roedy Green, Canadian Mind Products
 * @version 1.1 2011-01-25 allow you to specify encoding
 * @since 2010-12-14
 */
public final class LinesToCSV
    {
    // ------------------------------ CONSTANTS ------------------------------

    /**
     * how to use the command line
     */
    private static final String USAGE = "LinesToCSV needs a single filename.txt on the command line following by the grouping number.";

    // -------------------------- PUBLIC INSTANCE  METHODS --------------------------

    /**
     * convert a txt file to CSV, constructor. Just create. There are no methods to call.
     *
     * @param file          CSV file to be packed to remove excess space and quotes.
     * @param grouping      how many lines to collect together as one csv line, one field per line.
     * @param separatorChar field separator character, usually ',' in North America, ';' in Europe and sometimes '\t' for
     *                      tab for the output file.  It is tab for the input file.
     *                      Note this is a 'char' not a "string".
     * @param quoteChar     character used to quote fields containing awkward chars.
     * @param commentChar   char to use to introduce comments.  Use (char) 0 if none.  Only one character allowed.
     *                      Note this is a 'char' not a "string". For output file. No comment presumed in input file.
     * @param encoding      encoding of input and output file.
     *
     * @throws java.io.IOException if problems reading/writing file
     */
    @SuppressWarnings( { "WeakerAccess" } )
    public LinesToCSV( final File file, final int grouping, final char separatorChar, final char quoteChar, final char commentChar, final String encoding ) throws IOException
        {
        // O P E N

        final FileInputStream fis = new FileInputStream( file );
        final BufferedInputStream bis = new BufferedInputStream( fis, 32768 /* 32K bytes, 25% of allocation is optimal */ );
        final InputStreamReader eisr = new InputStreamReader( bis, "UTF-8" );
        final BufferedReader r = new BufferedReader( eisr, 24576 /* 24K chars (48K bytes), 75% of allocation is optimal */ );

        String outFilename = file.getAbsolutePath();
        int place = outFilename.lastIndexOf( '.' );
        if ( place < 0 )
            {
            outFilename = outFilename + ".csv";
            }
        else
            {
            outFilename = outFilename.substring( 0, place ) + ".csv";
            }

        final File outFile = new File( outFilename );
        // writer, quoteLevel, separatorChar, quoteChar, commentChar, trim
        final PrintWriter pw = new PrintWriter( new OutputStreamWriter( new BufferedOutputStream( new FileOutputStream( outFile ), 32768 ), encoding ) );
        // writer, quoteLevel, separatorChar, quoteChar, commentChar, trim// writer, quoteLevel, separatorChar, quoteChar, commentChar, trim
        final CSVWriter w = new CSVWriter( pw, 0 /* minimal  */, separatorChar, quoteChar, commentChar, true );
        int rCount = 0;
        int count = grouping;
        try
            {
            String thisLine;
            while ( ( thisLine = r.readLine() ) != null )
                {
                rCount++;
                w.put( thisLine );
                count--;
                if ( count == 0 )
                    {
                    w.nl();
                    count = grouping;
                    }
                }
            }
        catch ( EOFException e )
            {
            }
        finally
            {
            out.println( rCount + " text lines read, "
                         + w.getLineCount() + " csv lines written" );

            r.close();
            w.close();
            }
        }

    // --------------------------- main() method ---------------------------

    /**
     * Simple command line interface to LinesToCSV.  Converts text file of one field per line to a comma-separated CSV file.
     * java.exe com.mindprod.LinesToCSV somefile.txt  3
     * How many lines should be grouped together on one line of the CSV file.
     * Output is somefile.csv.  somefile.txt is left unchanged.
     * You can use LinesToCSV constructor in your own programs.
     *
     * @param args name of csv file to remove excess quotes and space
     */
    public static void main( String[] args )
        {
        if ( args.length != 2 )
            {
            throw new IllegalArgumentException( USAGE );
            }
        String filename = args[ 0 ];
        int grouping = Integer.parseInt( args[ 1 ] );
        if ( !filename.endsWith( ".txt" ) )
            {
            throw new IllegalArgumentException( "Bad Extension\n" + USAGE );
            }
        try
            {
            // file,  separatorChar,  quoteChar,  commentChar
            new LinesToCSV( new File( filename ), grouping, ',', '\"', '#', "UTF-8" );
            }
        catch ( IOException e )
            {
            err.println();
            e.printStackTrace( err );
            err.println( "LinesToCSV failed to convert " + filename );
            err.println();
            }
        }// end main
    }
