/*
 * [Chase.java]
 *
 * Summary: Chase redirection to find where page permanently redirected if any.
 *
 * Copyright: (c) 1998-2011 Roedy Green, Canadian Mind Products, http://mindprod.com
 *
 * Licence: This software may be copied and used freely for any purpose but military.
 *          http://mindprod.com/contact/nonmil.html
 *
 * Requires: JDK 1.5+
 *
 * Created with: JetBrains IntelliJ IDEA IDE http://www.jetbrains.com/idea/
 *
 * Version History:
 *  1.9 2008-08-22 support accept-charset, accept-encoding and accept-language. Fix bugs in gzip support.
 *  2.0 2009-02-20 major refactoring. separate setParms and setPostParms. new send method. Post can have both types of parm.
 *  2.1 2010-02-07 new methods Post.setBody Http.setRequestProperties.
 *  2.2 2010-04-05 new method getURL
 *  2.3 2010-11-14 new method setInstanceFollowRedirects
 */
package com.mindprod.http;

import java.io.IOException;
import java.net.HttpURLConnection;
import java.net.URL;

import static java.lang.System.out;

/**
 * Chase redirection to find where page permanently redirected if any.
 *
 * @author Roedy Green, Canadian Mind Products
 * @version 2.3 2010-11-14 new method setInstanceFollowRedirects
 * @since 2010
 */
@SuppressWarnings( { "WeakerAccess" } )
public final class Chase extends Http
    {
    // -------------------------- PUBLIC INSTANCE  METHODS --------------------------

    /**
     * constructor
     */
    public Chase()
        {
        }

    /**
     * Chase redirection to find where page permanently redirected if any.
     *
     * @param url complete URL including any parms, pre-encoded (use Http.encodeParms).
     *            might be http: or https:.
     *
     * @return host's response where URL permanently redirected, null if not redirected.
     * @see Get#send
     */
    @SuppressWarnings( { "UnusedAssignment", "MethodNamesDifferingOnlyByCase" } )
    public String send( final URL url )
        {
        try
            {
            // defaults
            responseCode = DEFAULT_RESPONSE_CODE;
            responseMessage = DEFAULT_RESPONSE_MESSAGE;
            this.url = url;
            // O P E N
            // will return HttpURLConnection for http: or HttpsURLConnection for https:
            final HttpURLConnection urlc = ( HttpURLConnection ) url.openConnection();
            urlc.setAllowUserInteraction( false );
            urlc.setDoInput( true ); // needed to see Location
            urlc.setDoOutput( false );// nothing beyond original request
            urlc.setUseCaches( false );
            urlc.setRequestMethod( "GET" );
            setStandardProperties( urlc );
            urlc.setInstanceFollowRedirects( false ); // we chase only the first leg. Otherwise we won't know about redirects.
            if ( DEBUGGING )
                {
                dumpHeaders( "---sent headers---", urlc );
                }
            // this is similar to HTTP.processResponse, except we don't read the page, and return the location.
            // send the message. Won't return from connect until other end responds with header.
            urlc.connect();
            // getInputStream not needed.
            // getResponseCode will block until the server responds.
            // save responseCode for later retrieval
            responseCode = urlc.getResponseCode();
            responseMessage = urlc.getResponseMessage();
            // only HTTP 301 permanent redirects and OK count as redirects, not temporary
            if ( DEBUGGING )
                {
                dumpHeaders( "---response headers---", urlc );
                }
            // do longhand so can trace in debugger
            final String location;
            if ( 301 <= responseCode && responseCode <= 303 )
                {
                location = urlc.getHeaderField( "Location" );
                }
            else
                {
                location = null;
                }
            // don't bother to read page content or even open the InputStream.
            if ( DEBUGGING )
                {
                out.println( "--------------------------------" );
                out.println( "ResponseCode:" + responseCode );
                out.println( "ResponseMessage:" + responseMessage );
                out.println( "Location:" + location );
                }
            // C L O S E
            urlc.disconnect();
            return location;
            }
        catch ( ClassCastException e )
            {
            // was not an http: url
            return null;
            }
        catch ( IOException e )
            {
            return null;
            }
        }// end get
    }
