/*
 * [Probe.java]
 *
 * Summary: simulates a browser posting a form to CGI via HEAD/GET.
 *
 * Copyright: (c) 1998-2011 Roedy Green, Canadian Mind Products, http://mindprod.com
 *
 * Licence: This software may be copied and used freely for any purpose but military.
 *          http://mindprod.com/contact/nonmil.html
 *
 * Requires: JDK 1.5+
 *
 * Created with: JetBrains IntelliJ IDEA IDE http://www.jetbrains.com/idea/
 *
 * Version History:
 *  2.0 2009-02-20 major refactoring. separate setParms and setPostParms. new send method. Post can have both types of parm.
 *  2.1 2010-02-07 new methods Post.setBody Http.setRequestProperties.
 *  2.2 2010-04-05 new method getURL
 *  2.3 2010-11-14 new method setInstanceFollowRedirects
 */
package com.mindprod.http;

import java.io.IOException;
import java.io.InputStream;
import java.net.HttpURLConnection;
import java.net.URI;
import java.net.URISyntaxException;
import java.net.URL;

/**
 * simulates a browser posting a form to CGI via HEAD/GET.
 * <p/>
 * It does not read any data. It is used for link checking where the server might ignore HEAD requests.
 *
 * @author Roedy Green, Canadian Mind Products
 * @version 2.3 2010-11-14 new method setInstanceFollowRedirects
 * @since 1998
 */
@SuppressWarnings( { "WeakerAccess" } )
public final class Probe extends Http
    {
    // -------------------------- PUBLIC INSTANCE  METHODS --------------------------

    /**
     * constructor
     */
    public Probe()
        {
        }

    /**
     * Check if a URL is working. Does a HEAD probe, if that fails does a GET probe in case server is ignoring HEAD requests.
     * We don't pay any attention to the content.
     *
     * @param url complete URL including get parm,  http: or https:
     *
     * @return responseCode
     */
    @SuppressWarnings( { "UnusedAssignment", "MethodNamesDifferingOnlyByCase" } )
    public int send( URL url )
        {
        try
            {
            // defaults
            responseCode = DEFAULT_RESPONSE_CODE;
            responseMessage = DEFAULT_RESPONSE_MESSAGE;
            this.url = url;
            // O P E N
            // will return HttpURLConnection for http: or HttpsURLConnection for https:
            HttpURLConnection urlc = ( HttpURLConnection ) url.openConnection();
            urlc.setAllowUserInteraction( false );
            urlc.setDoInput( true );
            urlc.setDoOutput( false );// nothing beyond original request
            urlc.setUseCaches( false );
            // use a head first, it is the official way.
            urlc.setRequestMethod( "HEAD" );
            setStandardProperties( urlc );
            urlc.connect();
            // getResponseCode will block until the server responds.
            // save responseCode for later retrieval
            responseCode = urlc.getResponseCode();
            responseMessage = urlc.getResponseMessage();
            urlc.disconnect();
            if ( responseCode == HttpURLConnection.HTTP_OK /* 200 */ )
                {
                return responseCode;
                }
            // HEAD failed. It may be because the server is being a jerk and ignoring HEADs. Try a GET.
            // get a fresh connection
            // will return HttpURLConnection for http: or HttpsURLConnection for https:
            urlc = ( HttpURLConnection ) url.openConnection();
            urlc.setAllowUserInteraction( false );
            urlc.setDoInput( true );
            urlc.setDoOutput( false );// nothing beyond original request
            urlc.setUseCaches( false );
            // use a head first, it is the official way.
            urlc.setRequestMethod( "GET" );
            setStandardProperties( urlc );
            urlc.connect();
            // getResponseCode will block until the server responds.
            // save responseCode for later retrieval
            responseCode = urlc.getResponseCode();
            responseMessage = urlc.getResponseMessage();
            // we need to be polite and dispose of the InputStream, even if we don't read the response.
            final InputStream is = urlc.getInputStream();
            // C L O S E
            is.close();
            urlc.disconnect();
            return responseCode;
            }
        catch ( ClassCastException e )
            {
            // was not an http: url
            return responseCode;
            }
        catch ( IOException e )
            {
            return responseCode;
            }
        }// end probe

    /**
     * Send a form full of data to the CGI host using HEAD/GET.
     * Must do a setParms beforehand. Lets you know if page exists.
     *
     * @param host   host name of the website, Should be form:"mindprod.com", no lead http://.
     * @param port   -1 if default, 8081 for local echoserver.
     * @param action action of form, page on website. Usually has a lead /.
     *
     * @return responseCode
     */
    @SuppressWarnings( { "UnusedAssignment", "MethodNamesDifferingOnlyByCase" } )
    public int send( String host, int port, String action )
        {
        try
            {
            // defaults
            responseCode = DEFAULT_RESPONSE_CODE;
            responseMessage = DEFAULT_RESPONSE_MESSAGE;
            // O P E N
            // URL will encode target and parms.
            URL url = new URI( "http",
                    null,
                    host,
                    port,
                    action,
                    null,
                    null ).toURL();
            // encoding does not matter since we are not interested in text of result, just status code.
            url = new URL( url.toString() + getEncodedParms( "UTF-8" ) );
            return send( url );
            }
        catch ( URISyntaxException e )
            {
            return responseCode;
            }
        catch ( IOException e )
            {
            return responseCode;
            }
        }// end probe
    }
